package wwc.messaging;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.Socket;
import java.util.Vector;

import salsa.language.Actor;
import salsa.language.ActorReference;
import salsa.language.Message;
import salsa.language.ServiceFactory;
import salsa.messaging.TransportService;
import salsa.naming.UAL;
import salsa.naming.UAN;

/**
 * This class is the reference implementation of the MessagingService. The
 * service is implemented as a RMSP client.
 * 
 * This implementation uses the NamingService interface and the salsa.language.
 * ServiceFactory for UAN resolution.
 * 
 * @author stepha
 */
public class Messenger implements TransportService {
	/**
	 * Constructor for Messenger.
	 */
	public Messenger() {
		super();
	}

	public void migrate(Actor state, UAL target) {
		Socket outgoing = null;
		try {
			outgoing = new Socket(target.getHost(), target.getPort());
			ObjectOutputStream outStream = new ObjectOutputStream(outgoing.getOutputStream());
			outStream.writeObject( state );
			outStream.flush();
		} catch (IOException e) {
			System.err.println("Migration of actor caused exception: ");
			System.err.println("\tBy actor: " + state.toString());
			System.err.println("\tTo target: " + target.toString());
			System.err.println("\tWith exception: " + e);
		}

		// Wait for the Acknowledgement that the actor is situated on the remote theater.
		try {
			RMSPAcknowledgement ack = (RMSPAcknowledgement) new ObjectInputStream( outgoing.getInputStream() ).readObject();
		} catch (Exception e) {
			System.err.println("Reception of RMSPAcknowledgement for actor migration caused exception: ");
			System.err.println("\tBy actor: " + state.toString());
			System.err.println("\tTo target: " + target.toString());
			System.err.println("\tWith exception: " + e);
		}

		try {
			outgoing.close();
		} catch (IOException e) {
			System.err.println("Migration of actor caused exception: ");
			System.err.println("\tBy actor: " + state.toString());
			System.err.println("\tTo target: " + target.toString());
			System.err.println("\tWith exception: " + e);
		}
	}

	public void send(Message message, UAL target) {
		// Determine the destination of this message.
		if (target == null) {
			UAN uan = message.getTarget().getUAN();

			if (uan != null) {
				Object entry = ServiceFactory.getTheater().getEntry(uan);
				if ((entry instanceof UAL) || (entry == null)) {
					// Resolve the location.
					target = ServiceFactory.getNaming().resolve( uan );
					ServiceFactory.getTheater().setEntry(uan, target);
				} else {
					((Actor)entry).send(message);
					return;
				}
			} else {
				System.err.println("Transport service error: ");
				System.err.println("\tSending message: " + message.getMethodName());
				if (message.getSender() != null) {
					System.err.println("\tBy actor: " + message.getSender().toString());
				} else System.err.println("\tBy actor: unknown");
				System.err.println("\tTo target: " + target.toString());
				System.err.println("\tUAN and UAL of message are null, cannot send message");
			}
		}

		try {
			Socket outgoing = new Socket(target.getHost(), target.getPort());
			ObjectOutputStream outStream = new ObjectOutputStream(outgoing.getOutputStream());
			outStream.writeObject( message );
			outStream.flush();
			outStream.close();
			outgoing.close();
		} catch (IOException e) {
			System.err.println("Transport service error: ");
			System.err.println("\tSending message: " + message.getDescriptor());
			if (message.getSender() != null) {
					System.err.println("\tBy actor: " + message.getSender().toString());
			} else System.err.println("\tBy actor: unknown");
			System.err.println("\tTo target: " + target.toString());
			System.err.println("\tException: " + e);
		}
	}

	public void send(Vector messages) {
		//used by the compiler to send a vector of messages generated by a message

		for (int i = 0; i < messages.size(); i++) {
			final Message current = (Message)messages.get(i);

			final ActorReference target = current.getTarget();
			try {
				if (current.getProperty() != null && current.getProperty().equals("delay")) {
					//This supports sending messages with the delay property
					Object[] parameters = current.getPropertyParameters();
					final long sleepTime = ((Integer)parameters[0]).intValue();

					Runnable delayThread = new Runnable() {
						public void run() {
							try {
								Thread.sleep(sleepTime);
							} catch (Exception e) {
								System.err.println("Error in delay thread for message: " + current.getName());
							}
							target.send(current);
						}
					};
					new Thread(delayThread, "Delay Thread").start();
				} else {
					target.send( current );
				}
			} catch (Exception e) {
				System.err.println("Transport service error: ");
				System.err.println("\tSending message: " + current.getDescriptor());
				if (current.getSender() != null) {
						System.err.println("\tBy actor: " + current.getSender().toString());
				} else System.err.println("\tBy actor: unknown");
				System.err.println("\tTo target: " + current.toString());
				System.err.println("\tException: " + e);
			}
		}
	}
}
