package wwc.messaging; 

import java.io.IOException;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.UnknownHostException;
import java.util.Hashtable;
import java.util.PropertyResourceBundle;

import salsa.language.Actor;
import salsa.language.ServiceFactory;
import salsa.messaging.ReceptionService;
import salsa.messaging.TheaterService;
import salsa.naming.UAL;
import salsa.naming.UAN;


/**
 * The Theater is the reference implementation of the TheaterService
 * interface.
 *
 * @version $Id: Theater.java,v 1.30 2003/10/09 14:14:04 deselt Exp $
 *
 * @author stepha
 */

public class Theater implements TheaterService {
	private static Hashtable registry = new Hashtable();
	private static Hashtable anonymousRegistry = new Hashtable();
	
	public static String version;
	
	private ServerSocket server;

	private static int port;
	public int getPort() 		{ return Theater.port; }
	public void setPort(int _port)	{ Theater.port = _port; }

	private boolean applet = false;
	public void isApplet() { applet = true; }

	private static int generatedTokens;
	private static int generatedLocations;

	/**
	 * Constructor for Theater.
	 */
	public Theater() {
		version  = "$Id: Theater.java,v 1.30 2003/10/09 14:14:04 deselt Exp $";
		server = null;
		Theater.port = 0;
		generatedLocations = 0;
		generatedTokens = 0;
	}

	/**
	 * These methods provide a way to generate a unique identifier
	 * for every token used by a SALSA program and a default UAL for
	 * new actors in the theater.
	 */
	private String rmspString = null;
	public String getLocation() {
		if (rmspString != null) {
			return rmspString;
		} else {
			try {
				rmspString = "rmsp://" + InetAddress.getLocalHost().getHostAddress() + ":" + Theater.port + "/";
			} catch (UnknownHostException e) {
				rmspString = "rmsp://localhost:" + Theater.port + "/";
			}
			return rmspString;
		}
	}
	public synchronized String getUniqueTokenId() {
		String generated = null;

		if (Theater.port == 0) {
			startTheater(null);
		}
		return getLocation() + "system/token" + (++Theater.generatedTokens) + "_" + System.currentTimeMillis();
	}
	public synchronized UAL generateUAL() {
		UAL generated = null;

		if (Theater.port == 0) {
			startTheater(null);
		}
		return new UAL(getLocation() + "system/location" + (++generatedLocations) + "_" + System.currentTimeMillis());
	}

	/**
	 * The following adds default access to the registries
	 * to allow for message sending and actor name resolution
	 */
	public Object getEntry(UAL ual) {
		synchronized( Theater.anonymousRegistry ) {
			return Theater.anonymousRegistry.get(ual.toString()); 
		}
	}
	public void setEntry(UAL ual, Object entry) {
		synchronized( Theater.anonymousRegistry ) {
			Theater.anonymousRegistry.put(ual.toString(), entry);
		}
	}
	public void removeEntry(UAL ual) {
		synchronized( Theater.anonymousRegistry ) {
			Theater.anonymousRegistry.remove(ual.toString());
		}
	}
	public Object getEntry(UAN uan) {
		synchronized( Theater.registry ) {
			return Theater.registry.get(uan.toString());
		}
	}
	public void setEntry(UAN uan, Object entry) {
		synchronized( Theater.registry ) {
			Theater.registry.put(uan.toString(), entry);
		}
	}
	public void removeEntry(UAN uan) {
		synchronized( Theater.registry ) {
			Theater.registry.remove(uan.toString());
		}
	}

	/**
	 * @see salsa.language.TheaterService#startTheater(int, salsa.language.ActorReference, java.lang.String)
	 */
	public TheaterService startTheater(Actor initial) {
		if (server == null) {
			//Initialize the port
			String propertyPort = null;
			if (!applet) propertyPort = System.getProperty("port");
			if (propertyPort != null) Theater.port = Integer.parseInt(propertyPort);

			// Create a server socket.
			try {
				server = new ServerSocket( Theater.port );
			} catch (IOException e) {
				System.err.println("Theater Service error:");
				System.err.println("\tCould not start theater.");
				System.err.println("\tException: " + e);
			}
			
			// Check to see what the port actually is.
			if (Theater.port == 0) {
				Theater.port = server.getLocalPort();
			}
			
			final TheaterService theater = this;
			final ServerSocket passive = server;
			
			// Accept connections.
			Runnable listenLoop = new Runnable() {
				public void run() {
					int i = 0;
					while (true) {
						try {
							final Socket incoming = passive.accept();

							Runnable handlerThread = new Runnable() {
								public void run() {
									ReceptionService handler = ServiceFactory.getReception();
									handler.process(incoming);
								}
							};
							new Thread(handlerThread, "Handler Loop " + i).start();
						} catch (IOException e) {
							System.err.println("Theater Service error: ");
							System.err.println("\tFailed to accept connection.");
							System.err.println("\teException: " + e);
						}

						i++;
					}
				}
			};
	
			// Start the listening thread.
			new Thread( listenLoop, "Listen Loop Thread" ).start();
		}

		// Add the specified actor to the theater.
		if (initial != null) setEntry( initial.getUAN(), initial );

		return this;
	}

	public static void main( String[] args ) {
		Thread.currentThread().setName( "Main startup thread.");
		
		// Check arguments.
		if (args.length > 1) {
			System.out.println( "WWC Theater Version: " + version);
			System.out.println( "Usage: java wwc.theater.Theater [<port>]" );
			System.out.println( "or invoke a Actor with VM option -Dtheater=wwc.theater.Theater and -Dport=<port>" );
			return;
		}
		
		// Create a theater.
		Theater theater = new Theater();
		
		// Determine port.
		Theater.port = 0;
		if (args.length == 1) {
			try {
				Theater.port = Integer.parseInt( args[0] );
			} catch (NumberFormatException e) {
				System.err.println( "Theater Service error:" );
				System.out.println( "\tBad port: " + args[0] );
				return;
			}
		}
		
		// Construct service factory.
		ServiceFactory.setTheater( theater );
		
		// Start the theater.
		theater.startTheater(null);

		try {
			String stringHere = "rmsp://" + InetAddress.getLocalHost().getHostAddress() + ":" + Theater.port;
			if ( System.getProperty("silent") == null ) System.err.println("Theater started at: " + stringHere);
		} catch (UnknownHostException e) {
			System.err.println("Theater Service error: ");
			System.err.println("\tUnknown local host");
			System.err.println("\tException: " + e);
		}
	}
}
