package wwc.naming;

import java.net.Socket;
import java.io.BufferedReader;
import java.io.PrintWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.StringTokenizer;


import salsa.language.exceptions.SalsaException;
import salsa.language.exceptions.SalsaWrapperException;
import salsa.naming.NamingService;
import salsa.naming.UAL;
import salsa.naming.UAN;

/**
 * This class implements a UANP client. After the class is constructed, several
 * threads may pass through it to conduct UANP operations through the
 * NamingService interface in salsa.language.
 * 
 * @version $Id: WwcNamingImpl.java,v 1.3 2003/09/15 06:46:16 deselt Exp $
 * @author Abe Stephens -- stepha@rpi.edu
 * @author Carlos Varela 
 */
public class WwcNamingImpl implements NamingService {

	public WwcNamingImpl() { }

	/**
	 * This method provides UANP GET functionality.
	 * 
	 * @see salsa.language.NamingService#resolve(salsa.language.UAN)
	 */
	public UAL resolve(UAN name) {

		try {
			Socket s = new Socket(name.getHost(), name.getPort());
			
			BufferedReader in = new BufferedReader(new InputStreamReader(s.getInputStream()));
			PrintWriter out = new PrintWriter(s.getOutputStream(), true);
			out.println("GET " + name.getIdentifier() + " UANP/1.0\n" );

			String inputLine = in.readLine();
			if (inputLine == null) throw new SalsaException("UANP Server not responding.");

			StringTokenizer st = new StringTokenizer(inputLine);
			st.nextToken();
			
			String returnCode = new String(st.nextToken());
			if (!returnCode.equals("200"))
				throw new SalsaException( "UANP Error: " + inputLine );

			return new UAL((new StringTokenizer(in.readLine())).nextToken());
		} catch (IOException e) {
			
			throw new SalsaWrapperException( "UANP GET Failed: ", e );
		} 
	}

	/**
	 * This method provides UANP PUT functionality.
	 * 
	 * @see salsa.language.NamingService#add(salsa.language.UAN, salsa.language.UAL)
	 */
	public void add(UAN name, UAL location) {
		
		try {
			Socket socket = new Socket(name.getHost(), name.getPort());
			
			BufferedReader in = new BufferedReader(new InputStreamReader(socket.getInputStream()));
			PrintWriter out = new PrintWriter(socket.getOutputStream(), true);
			
			out.println("PUT " + name.getIdentifier() + " " + location + " UAN/1.0\n");
			out.flush();
			String status = in.readLine();
		
			String response = in.readLine();

			in.close();
			out.close();
			socket.close();
			
		} catch (IOException e) {
			
			throw new SalsaWrapperException( "UANP PUT Failed: ", e );
		}		
	}

	/**
	 * This method is equivlent to add( name, location );
	 * 
	 * @see salsa.language.NamingService#update(salsa.language.UAN, salsa.language.UAL)
	 */
	public void update(UAN name, UAL location) {
		
		add( name, location );
	}

	/**
	 * This method is not implemented.
	 * 
	 * @see salsa.language.NamingService#delete(salsa.language.UAN, salsa.language.UAL)
	 */
	public void delete(UAN name, UAL location) {
		
		throw new SalsaException( "This naming service does not implement delete." );
	}

}
